#ifndef UMBRELLA_SIGNATURE
#define UMBRELLA_SIGNATURE

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/security.h>
#include <linux/gfp.h>
#include <linux/linkage.h>
#include <linux/errno.h>
#include <linux/namei.h>
#include <linux/proc_fs.h>

#include "umb_bitvector.h"

#include <linux/crypto.h>
#include <asm/scatterlist.h>
#include "../mpi/mpi.h"


/* Size of the Bsign header. */ 
#define UMBRELLA_ELF_SIG_SECTION 0x80736967      /* ((0x80 << 24)|('s' << 16)|('i' << 8)|'g') */
#define UMBRELLA_ELF_SIG_SIZE 512 /* Maybe 1024 ?*/
#define UMBRELLA_ELF_READ_BLOCK_SIZE 1024	/* Signature will be done in chunks of n bytes */

/* Internal structure of Umbrella Header */
#define UMBRELLA_SIG_START 0
#define UMBRELLA_ID_START 17
#define UMBRELLA_LV1_START 84
#define UMBRELLA_LV2_START 280
#define UMBRELLA_SIG_END 2332

#define UMBRELLA_SIG_SIZE 2346 /* Size of the Umbrella Header */
#define UMBRELLA_SIG_NO_LV2 128
#define UMBRELLA_ID_SIZE 3
#define UMB_KEY_SIZE_SIZE 3
#define UMB_KEY_ENDTAG_SIZE 11
#define UMB_KEY_STARTTAG_SIZE 13
#define UMB_KEY_ID_SIZE 64



/* RSA ting */
#define TVMEMSIZE	4096

#define UMBRELLA_PKEY_N_OFFSET  8	/* offset for pkey->n */

/* 
 * Format of digital signature done by bsign:
 * - "#1; bsign v%s\n"
 * - hash of file with sig section zerod (crypto class and timestamp not added to hash)
 * - length of umbrella (2 bytes)
 * - umbrella
 */

#define UMBRELLA_BSIGN_VERSION    "0.4.5"
#define UMBRELLA_BSIGN_STRING     "#1; bsign v" UMBRELLA_BSIGN_VERSION "\n"
#define UMBRELLA_BSIGN_GREET_SIZE sizeof(UMBRELLA_BSIGN_STRING)-1
#define UMBRELLA_BSIGN_HASH       20	/* sha1 hash */
#define UMBRELLA_BSIGN_LEN_OFFSET 2	/* length of umbrella added by bsign */
#define UMBRELLA_BSIGN_INFOS      UMBRELLA_BSIGN_GREET_SIZE+UMBRELLA_BSIGN_HASH+UMBRELLA_BSIGN_LEN_OFFSET

/* GPG .sig file/section structure:
 * 89     00 95 03          05      00    3F293CEB  
 * hdrlen       sig version md5 len class timestamp 
 * 
 * D39C2077 C307562E 01          02          29
 * keyid[0] keyid[1] pubkey algo digest algo digest start[0]
 * 
 * C2              03FF         Rest of .sig
 * digest start[1] nbits of MPI MPI (as read by mpi_read)
 */

#define UMBRELLA_RSA_CLASS_OFFSET     5
#define UMBRELLA_RSA_TIMESTAMP_OFFSET 6
#define UMBRELLA_RSA_DATA_OFFSET      22


/*ToDO: this is a constraint, we suppose that the max size of a
  big integer is 1024 bytes. This needs to be modified in order to
  have a dynamic way of allocating memory. */

#define UMBRELLA_MPI_MAX_SIZE_N 1024
#define UMBRELLA_MPI_MAX_SIZE_E 128

/**
 * Supported algorithms
 */
#define HASH_SHA1 0
#define SIGN_RSA 0

typedef struct sig_ctx_st {
	struct crypto_tfm *tfm;
	struct scatterlist sg[1];
	int digestAlgo;
	int signAlgo;
	char *tvmem;
} SIGCTX;

struct dsig {
	char *bsig;
	char *usig;
	char *key;
	char *keyid;
};

struct parsed_sig {
	char *sig_start;
	char *id;
	bitvector lv1;
	char **lv2;
};


/* Entry point */
int umb_handle_signature(struct linux_binprm * bprm);

int umb_sign_verify_final(SIGCTX * ctx, char *sig, int siglen /* PublicKey */ ,
						  unsigned char *signed_hash, char *key_id);

int umb_sign_verify_update(SIGCTX * ctx, char *buf, int buflen);

SIGCTX *umb_sign_verify_init(int hashalgo, int signalgo);

int umb_verify_signature(Elf32_Shdr * elf_shdata, char *sig_orig, 
						 struct file *file, int sh_offset, char *key_id);

int umb_get_pkey(unsigned char *raw_public_key_n,
				 unsigned char *raw_public_key_e, char *pkey_file);

int umb_get_pkey_from_mem(unsigned char *, unsigned char *, char *);

struct parsed_sig* umb_parse_usig(struct linux_binprm *bprm, char* usig  );
 
int umb_verify_elf(struct elfhdr *elf_hdr);

struct elfhdr *umb_read_elf_header(struct file *file);

Elf32_Shdr *umb_read_section_header(struct file *file, int sh_size, int sh_off);

struct dsig* umb_find_signature(struct elfhdr *elf_hdr, Elf32_Shdr * elf_shdata, 
								struct file *file,int *sh_offset);

int umb_init_pkey(const char read_par, unsigned char *raw_public_key, int mpi_size);


int umb_check_signature(Elf32_Shdr *elf_shdata, struct dsig *sig, 
						struct file *file, int sh_offset);

void umb_free_sig(struct parsed_sig* usig);


int umb_cached_signature(struct linux_binprm *bprm);

/*local helper function */
void umb_test_sig(char* usig);

/* Set the restrictions to the correct pid */

void umb_apply_policy(struct linux_binprm * bprm);

struct parsed_sig* parsed_sig_create(void);




#endif /*UMBRELLA_SIGNATURE*/
