/******************************************************************************/
/* $Header: /cvsroot/umbrella/umbrella-devel/kernel/security/umbrella/umb_fsr.c,v 1.5 2005/04/28 11:45:26 ipqw Exp $ */
/******************************************************************************/
/*  AUTHOR:                                                                   */
/******************************************************************************/
/*                                                                            */
/*    Sren Nhr Christensen                                                  */
/*    Kristian Srensen                                                       */
/*    Michel Thryse                                                          */
/*    mailto:umbrella@cs.auc.dk                                               */
/*    http://umbrella.sourceforge.net                                         */
/*                                                                            */
/******************************************************************************/
/*  COPYRIGHT:                                                                */
/******************************************************************************/
/*                                                                            */
/*    Copyright (c) 2005 The Umbrella Security Team.                          */
/*    All rights reserved.                                                    */
/*                                                                            */
/******************************************************************************/
/*  LICENSE:                                                                  */
/******************************************************************************/
/*                                                                            */
/*    This file is part of the Umbrella Security Project and is free          */
/*    software; you can redistribute it and/or modify it under the terms      */
/*    of the GNU Library General Public License as published by the Free      */
/*    Software Foundation;                                                    */
/*    The code is distributed in the hope that it will be useful,             */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        */
/*    Library General Public License for more details.                        */
/*                                                                            */
/*    You should have received a copy of the GNU Library General Public       */
/*    License along with this library; if not, write to the                   */
/*    Free Software Foundation, Inc.,                                         */
/*    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   */
/*                                                                            */
/*    or download a copy from ftp://ftp.gnu.org/pub/gnu/COPYING.LIB-2.0       */
/*                                                                            */
/******************************************************************************/


#include <linux/kernel.h>

#include "include/umb_types.h"
#include "include/umb_fsr.h"



void fsr_free(struct fsr *root) {

	int i = 0;

	if (root) {
		if (root->successors != NULL)
			while (root->successors[i] != NULL)
				fsr_free(root->successors[i++]);
		
		
		kfree(root->successors);
		kfree(root->name);
		kfree(root);
	}

}



void fsr_print_dot(const struct fsr *root) {

	int i = 0;

	if (root->successors != NULL)
		while (root->successors[i] != NULL) {
			fsr_print(root->successors[i]);
			printk("\t\"%s\" -> \"%s\";\n", root->name,
				   root->successors[i]->name);
			i++;
		}

}



void fsr_print(const struct fsr *root) {

	printk("digraph FSR\n{\n");
	fsr_print_dot(root);
	printk("}\n");

}



int fsr_check(struct fsr *root, char **path) {
	int succ_depth, path_depth;
	struct fsr *current_node;

	if (!root || !path)
		return ALLOWED;

	succ_depth = 0;
	path_depth = 0;

	/* traverse tree while finding matching nodes */
	current_node = root->successors[succ_depth];
	while (current_node) {
		if (!path[path_depth]) {
			printk("Error: Path element NULL\n");
			printk("breaking\n");
			break;
		}

		/* printk("comparing %s and %s\n", current_node->name, path[path_depth]); */
		if (!strcmp(current_node->name, path[path_depth])) {
			succ_depth = 0;
			path_depth++;
			root = current_node; /* variable reuse, locally only */
			current_node = root->successors[succ_depth];

			if (path[path_depth] && current_node) {
				/* printk("path matching so far\n"); */
				continue;
			}
			else if (!path[path_depth] && current_node) {
				/* printk("path not in tree -- too short (ALLOW)\n"); */
				return ALLOWED;
			}
			else if (path[path_depth] && !current_node) {
				/* printk("restriction more general than path (DENY)\n"); */
				return DENIED;
			}
			else if (!path[path_depth] && !current_node) {
				/* printk("path in tree -- exact match (DENY)\n"); */
				return DENIED;
			} else {
				printk("Error!!\n");
			}
		}
		else {
			succ_depth++;
			current_node = root->successors[succ_depth];
		}
	}

	/* return allowed for all processes not restricted by anything */
	/* i.e. processes that only has the dummy node set by fsr_init */
	return ALLOWED;

}



struct fsr *fsr_clone(struct fsr *fsr) {

	struct fsr *new_fsr = kmalloc(sizeof(struct fsr), GFP_ATOMIC);
	struct fsr **succ = NULL;

	int len = strlen(fsr->name);
	char *name = kmalloc(sizeof(char) * ++len, GFP_ATOMIC);

	if (!new_fsr) {
		printk("Umbrells FSR error: Not enough memory!\n");
		return NULL;
	}

	strcpy(name, fsr->name);

	/*      printk("cloning %s\n", name); */

	new_fsr->name = name;
	new_fsr->successors = succ;

	return new_fsr;

}



int fsr_copy(struct fsr *root_dst, struct fsr *root_src) {

	int count;

	/* stop if any of the arguments is null */
	if (!root_dst || !root_src)
	    return 0;

	/* stop the recursion when leaf is reached */
	if (!root_src->successors[0]) {
		root_dst->successors = kmalloc(sizeof(struct fsr *), GFP_ATOMIC);

		if (!root_dst->successors) {
			printk("Umbrells FSR error: Not enough memory!\n");
			return 0;
		}

		root_dst->successors[0] = NULL;
		return 0;
	}

	/* count successors */
	count = 0;
	while (root_src->successors[count])
		count++;

	/* allocate memory for successors */
	count++; /* still need space for the tailing NULL */
	root_dst->successors = kmalloc(sizeof(struct fsr *) * count, GFP_ATOMIC);

	if (!root_dst->successors) {
		printk("Umbrells FSR error: Not enough memory!\n");
		return 0;
	}


	/* copy the successors */
	count = 0;
	while (root_src->successors[count]) {
		root_dst->successors[count] =
			fsr_clone(root_src->successors[count]);
		count++;
	}
	root_dst->successors[count] = NULL;

	/* traverse sub tree */
	count = 0;
	while (root_src->successors[count]) {
		fsr_copy(root_dst->successors[count], root_src->successors[count]);
		count++;
	}

	return 0;

}


/* The first while loop traverses the the tree, finding nodes matching elements
 * of path. When a matching element is found, if the path is empty, but a sub-tree
 * exists, this is removed.  When no more matching nodes if found in the tree, the
 * second while loop inserts the remaining nodes for the path */
int fsr_insert(struct fsr *root, char **path) {

	int counter, succ_depth, path_depth, i;
	struct fsr *current_node, *new_node;
	struct fsr **new_succ, **succ;

	succ_depth = 0;
	path_depth = 0;

	/* traverse tree while finding matching nodes */
	current_node = root->successors[succ_depth];
	while (current_node) {
		if (!path[path_depth]) {
			printk("Error: Path element NULL\n");
			printk("breaking\n");
			break;
		}

		/*              printk("comparing %s and %s\n", current_node->name, path[path_depth]); */
		if (!strcmp(current_node->name, path[path_depth])) {
			succ_depth = 0;
			path_depth++;
			/* inserting more restrictive restriction => free subtree(s) */
			if (!path[path_depth]) {
				i = 0;
				while (current_node->successors[i]) {
					/*                                      printk("freeing %s subtree\n", current_node->successors[i]->name); */
					fsr_free(current_node->successors[i]);
					current_node->successors[i] = NULL;
					i++;
				}
			}
			/*                      printk("root is %s\n", root->name); */
			root = current_node; /* variable reuse locally: root is used later! */
			current_node = root->successors[succ_depth];
			continue;
		}
		else {
			succ_depth++;
			current_node = root->successors[succ_depth];
		}
	}

	/*      printk("no (more) matches found\n"); */

	/* insert remaining path elements */
	while (path[path_depth]) {
		/*              printk("inserting %s\n", path[path_depth]); */

		/* count how many successors are present */
		counter = 0;
		if (current_node)
			while (current_node->successors[counter])
				counter++;
		else {
			current_node = root;
			while (current_node->successors[counter])
				counter++;
		}

		/* increase successors by 1 to make room for new one */
		/* + 1 extra for the tailing NULL */
		counter = counter + 2;

		/* allocate memory */
		new_succ = kmalloc(sizeof(struct fsr *) * counter, GFP_ATOMIC);
		/* printk("allocated memory for %i successors\n", counter); */

		if (!new_succ) {
			printk("Umbrells FSR error: Not enough memory!\n");
			return 0;
		}


		/* copy "old" pointers */
		counter = 0;
		while (current_node->successors[counter]) {
			new_succ[counter] = current_node->successors[counter];
			counter++;
		}

		/* insert new element to new_succ */
		new_node = kmalloc(sizeof(struct fsr), GFP_ATOMIC);
		succ = kmalloc(sizeof(struct fsr *), GFP_ATOMIC);

		if (!new_node || !succ) {
			printk("Umbrells FSR error: Not enough memory!\n");
			return 0;
		}


		new_node->name = path[path_depth];
		new_node->successors = succ;
		new_node->successors[0] = NULL;

		new_succ[counter] = new_node;
		counter++;
		new_succ[counter] = NULL;

		/* swap new and old successors */
		kfree(current_node->successors);
		current_node->successors = new_succ;


		current_node = current_node->successors[counter-1];
		path_depth++;
	}

	return 0;

}



struct fsr *fsr_init() {

	struct fsr *dummy = kmalloc(sizeof(struct fsr), GFP_ATOMIC);
	struct fsr **successors = kmalloc(sizeof(struct fsr *), GFP_ATOMIC);

	if (!dummy || !successors) {
		printk("Umbrells FSR error: Not enough memory!\n");
		return 0;
	}

	dummy->name = (char *) kmalloc(sizeof(char)*6, GFP_ATOMIC);


	if (!dummy->name) {
		printk("Umbrells FSR error: Not enough memory!\n");
		return 0;
	}

	strcpy(dummy->name, "dummy");

	successors[0] = NULL;
	dummy->successors = successors;

	return dummy;

}
