/******************************************************************************/
/* $Header: /cvsroot/umbrella/umbrella-devel/kernel/security/umbrella/umb_signature.c,v 1.35 2005/03/31 13:28:38 mthrysoe Exp $ */
/******************************************************************************/
/*  AUTHOR:                                                                   */
/******************************************************************************/
/*                                                                            */
/*    Sren Nhr Christensen                                                  */
/*    Kristian Srensen                                                       */
/*    Michel Thryse                                                          */
/*    mailto:umbrella@cs.auc.dk                                               */
/*    http://umbrella.sourceforge.net                                         */
/*                                                                            */
/******************************************************************************/
/*  COPYRIGHT:                                                                */
/******************************************************************************/
/*                                                                            */
/*    Copyright (c) 2004 The Umbrella Security Team.                          */
/*    All rights reserved.                                                    */
/*                                                                            */
/******************************************************************************/
/*  LICENSE:                                                                  */
/******************************************************************************/
/*                                                                            */
/*    This file is part of the Umbrella Security Project and is free          */
/*    software; you can redistribute it and/or modify it under the terms      */
/*    of the GNU Library General Public License as published by the Free      */
/*    Software Foundation;                                                    */
/*    The code is distributed in the hope that it will be useful,             */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        */
/*    Library General Public License for more details.                        */
/*                                                                            */
/*    You should have received a copy of the GNU Library General Public       */
/*    License along with this library; if not, write to the                   */
/*    Free Software Foundation, Inc.,                                         */
/*    59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   */
/*                                                                            */
/*    or download a copy from ftp://ftp.gnu.org/pub/gnu/COPYING.LIB-2.0       */
/*                                                                            */
/******************************************************************************/
/*  COMMENTS: Original code from Digsig 1.3.1                                 */
/******************************************************************************/



/* RSA stuff **/
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/crypto.h>

#include "cipher/rsa-verify.h"
#include <asm/scatterlist.h>
#include "mpi/mpi.h"

/* End of RSA stuff */

#include "include/umb_signature.h"
#include "include/umb_bitvector.h"
#include "include/umb_types.h"
#include "include/umb_keyring.h"


static DECLARE_MUTEX(umbrella_sem);

int gDigestLength[] = { /* SHA-1 */ 0x14 };
static struct crypto_tfm *sha1_tfm;

/******************************************************************************
                             Internal functions 
******************************************************************************/

static int  umb_sha1_init(SIGCTX * ctx);
static void umb_sha1_update(SIGCTX * ctx, char *buf, int buflen);
static int umb_sha1_final(SIGCTX * ctx, char *digest);
static int umb_rsa_bsign_verify(unsigned char *hash_format, int length,
								unsigned char *signed_hash, char *key_id);


/******************************************************************************
Description : 
   Performs RSA verification of signature contained in binary
Parameters  : 
Return value: 0 - RSA signature is valid
              1 - RSA signature is invalid
              -1 - an error occured
******************************************************************************/

static int umb_rsa_bsign_verify(unsigned char *hash_format, int length,
								unsigned char *signed_hash, char *key_id) {
	int rc = 8;
	MPI hash = NULL;
	MPI data;
	unsigned nread = UMBRELLA_ELF_SIG_SIZE;
	int nframe;
	unsigned char sig_class;
	unsigned char sig_timestamp[SIZEOF_UNSIGNED_INT];
	int i;
	SIGCTX *ctx = NULL;
	unsigned char *new_sig;
	struct key *key = NULL;

	new_sig = kmalloc(gDigestLength[HASH_SHA1], GFP_KERNEL);

	if (!new_sig) {
		printk(KERN_INFO "Umbrella: kmalloc failed for new_sig\n");
		return -ENOMEM;
	}
	
	/* Get MPI of signed data from .sig file/section */
	nread = UMBRELLA_ELF_SIG_SIZE;

	data = mpi_read_from_buffer(signed_hash + UMBRELLA_RSA_DATA_OFFSET, 
				    &nread, 0);
	if (!data) {
		kfree(new_sig);
		return -EINVAL;
	}

	/* Get MPI for hash */
	/* bsign modif - file hash - gpg modif */
	/* bsign modif: add bsign greet at beginning */
	/* gpg modif:   add class and timestamp at end */

	ctx = kmalloc(sizeof(SIGCTX), GFP_KERNEL);
	if (!ctx) {
		printk(KERN_INFO "Umbrella: Cannot allocate ctx\n");
		mpi_free (data);
		kfree (new_sig);
		return -ENOMEM;
	}

	
	ctx->tvmem = kmalloc(TVMEMSIZE, GFP_KERNEL);
	if (!ctx->tvmem) {
		kfree (ctx);
		mpi_free(data);
		kfree (new_sig);
		printk(KERN_INFO "Umbrella: Cannot allocate plaintext buffer\n");
		return -ENOMEM;
	}

	umb_sha1_init(ctx);

	sig_class = signed_hash[UMBRELLA_RSA_CLASS_OFFSET];
	sig_class &= 0xff;

	for (i = 0; i < SIZEOF_UNSIGNED_INT; i++) {
		sig_timestamp[i] =
		    signed_hash[UMBRELLA_RSA_TIMESTAMP_OFFSET + i] & 0xff;
	}

	umb_sha1_update(ctx, UMBRELLA_BSIGN_STRING, UMBRELLA_BSIGN_GREET_SIZE);
	umb_sha1_update(ctx, hash_format, SHA1_DIGEST_LENGTH);
	umb_sha1_update(ctx, &sig_class, 1);
	umb_sha1_update(ctx, sig_timestamp, SIZEOF_UNSIGNED_INT);

 	if ((rc = umb_sha1_final(ctx, new_sig)) < 0) {
		printk(KERN_INFO "Umbrella: internal_rsa_verify_final: Cannot finalize hash algorithm\n");
		mpi_free(data); 
		kfree(ctx->tvmem);
		kfree(ctx);
		return rc; 
	}

	key = get_key(key_id);

	/* if no key was found, we will deny (rc=1) */
	if(!key) {
		rc = 1;
		goto out;
	}
	
	nframe = mpi_get_nbits(key->payload[0]);

	hash = do_encode_md(new_sig, nframe);

	if (hash == MPI_NULL) {
		printk(KERN_INFO "Umbrella: mpi creation failed!\n");
	}

	/* Do RSA verification */
	rc = rsa_verify(hash, &data, key->payload);

 out:
	mpi_free(hash);
	mpi_free(data);
	kfree(ctx->tvmem);
	kfree(ctx);
	kfree (new_sig);
	return rc;
}

/******************************************************************************
Description : This is where the RSA signature verification actually takes place
Parameters  : 
Return value: 
******************************************************************************/
int umb_sign_verify_final(SIGCTX * ctx, char *sig, int siglen /* PublicKey */ ,
						  unsigned char *signed_hash, char *key_id) {
	char *digest;
	int rc = -ENOMEM;

	digest = kmalloc (gDigestLength[ctx->digestAlgo], GFP_KERNEL);
	if (!digest) {
		printk(KERN_INFO "Umbrella: kmalloc failed for digest\n");
		kfree(digest);
		return rc;
	}
	/* TO DO: check the length of the signature: it should be equal to the length
	   of the modulus */
	if ((rc = umb_sha1_final(ctx, digest)) < 0) {
		printk(KERN_INFO "Umbrella: Cannot finalize hash algorithm\n");
		kfree(digest);
		return rc;
	}

	rc = -EINVAL;
	if (siglen < gDigestLength[ctx->digestAlgo]) {
		printk(KERN_INFO "Umbrella: signature is the WRONG length\n");
		kfree(digest);
		return rc;
	}
	memcpy(sig, digest, gDigestLength[ctx->digestAlgo]);

	switch (ctx->digestAlgo) {
	case SIGN_RSA:
		rc = umb_rsa_bsign_verify(digest, gDigestLength[ctx->digestAlgo],
								  signed_hash, key_id);
		
		break;
	default:
		printk(KERN_INFO "Umbrella: Unsupported cipher algorithm in binary digital signature verification\n");
	}
	return rc;

	/* free everything */
	/* do not free SHA1-TFM. For optimization, we choose always to use the same one */
}

/******************************************************************************
Description : 
Parameters  : 
Return value: 
******************************************************************************/


static int umb_sha1_final(SIGCTX * ctx, char *digest) {
	/* TO DO: check the length of the signature: it should be equal to the length
	   of the modulus */

	if (ctx == NULL)
		return -EINVAL;

	crypto_digest_final(ctx->tfm, digest);
	return 0;
}

/******************************************************************************
Description : Portability layer function. 
Parameters  : 
Return value: void. 
******************************************************************************/

static void umb_sha1_update(SIGCTX * ctx, char *buf, int buflen) {
	char *plaintext;

	memcpy(ctx->tvmem, buf, buflen);
	plaintext = (void *) ctx->tvmem;

	ctx->sg[0].page = virt_to_page(plaintext);
	ctx->sg[0].offset = ((long) plaintext & ~PAGE_MASK);
	ctx->sg[0].length = buflen;

	crypto_digest_update(ctx->tfm, ctx->sg, 1);
}

/******************************************************************************
Description : Portability layer function. 
Parameters  : 
Return value: int. 
******************************************************************************/

static int umb_sha1_init(SIGCTX * ctx) {
	if (ctx == NULL)
		return -1;

	if (sha1_tfm == NULL)
		sha1_tfm = crypto_alloc_tfm("sha1", 0);
	ctx->tfm = sha1_tfm;
	if (ctx->tfm == NULL) {
		printk(KERN_INFO "Umbrella: tfm allocation failed\n");
		return -1;
	}

	crypto_digest_init(ctx->tfm);
	return 0;
}

/******************************************************************************
Description : 
Parameters  : 
  ctx
  buf data to sign (part of a bigger buffer)
  buflen length of buf
Return value: 0 normally
******************************************************************************/
int umb_sign_verify_update(SIGCTX * ctx, char *buf, int buflen) {
	if (ctx == NULL)
		return -EINVAL;

	switch (ctx->digestAlgo) {
	case HASH_SHA1:
		umb_sha1_update(ctx, buf, buflen);
		break;
	default:
		printk(KERN_INFO "Umbrella: Unknown hash algo\n");
		return -EINVAL;
	}

	return 0;
}

/******************************************************************************
Description : 
Parameters  : 
  hashalgo the identifier of the hash algorithm to use (HASH_SHA1 for instance)
  signalgo sign algorithm identifier. Ex. SIGN_RSA
Return value: a signature context valid for this signature only
******************************************************************************/

SIGCTX *umb_sign_verify_init(int hashalgo, int signalgo) {

	SIGCTX *ctx;
	/* allocating signature context */

	ctx = kmalloc(sizeof(SIGCTX), GFP_KERNEL);
	if (!ctx) {
		printk(KERN_INFO "Umbrella: Cannot allocate ctx\n");
		return NULL;
	}

	ctx->tvmem = kmalloc(TVMEMSIZE, GFP_KERNEL);
	if (!ctx->tvmem) {
		printk(KERN_INFO "Umbrella: Cannot allocate plaintext buffer\n");
		return NULL;
	}

	switch (hashalgo) {
	case HASH_SHA1:
		if (umb_sha1_init(ctx)) {
			printk(KERN_INFO "Umbrella: Initializing SHA1 failed\n");
			if (ctx) {
				kfree(ctx->tvmem);
				kfree(ctx);
			}
			return NULL;
		}
		break;
	default:
		printk(KERN_INFO "Umbrella: Unknown hash algo\n");
		return NULL;
	}

	/* checking sign algo is known */
		switch (signalgo) {
	case SIGN_RSA:
		break;
	default:
		printk(KERN_INFO "Umbrella: Unknown sign algo\n");
		if (ctx) {
			kfree(ctx->tvmem);
			kfree(ctx);
		}

		return NULL;
	}
	ctx->digestAlgo = hashalgo;
	ctx->signAlgo = signalgo;
	return ctx;

}

/******************************************************************************
Description : verify if signature matches binary's signature
Parameters  : 
   elf_shdata is the data of the signature section
   sig_orig is the original signature of the binary
   file is the file handle of the binary
   sh_offset is offset of signature section in elf
Return value: 0 for false or 1 for true or -1 for error
******************************************************************************/
int umb_verify_signature(Elf32_Shdr * elf_shdata, char *sig_orig, 
						 struct file *file, int sh_offset, char *key_id) {
	char *sig_result = NULL, *read_blocks = NULL;
	int retval = -EPERM, offset;
	unsigned int lower, upper;
	loff_t i_size;
	SIGCTX *ctx = NULL;

	//printk("Umbrella: Entering function umb_verify_signature\n");
	down(&umbrella_sem);

	ctx = umb_sign_verify_init(HASH_SHA1, SIGN_RSA);
	if (!ctx ) {
		printk(KERN_INFO "Umbrella: Cannot allocate crypto context.\n");
		return -ENOMEM;
	}

	sig_result = kmalloc(UMBRELLA_ELF_SIG_SIZE, GFP_ATOMIC);
	if (!sig_result) {
		printk(KERN_INFO "Umbrella: kmalloc failed for sig_result.\n");
		return -ENOMEM;
	}

	read_blocks = kmalloc(UMBRELLA_ELF_READ_BLOCK_SIZE, GFP_KERNEL);
	if (!read_blocks) {
		printk(KERN_INFO "Umbrella: kmalloc failed for read_block.\n");
		return -ENOMEM;
	}

	memset(sig_result, 0, UMBRELLA_ELF_SIG_SIZE);

	i_size = i_size_read(file->f_dentry->d_inode);
	for (offset = 0; offset < i_size; offset += UMBRELLA_ELF_READ_BLOCK_SIZE){

		retval = kernel_read(file, offset, (char *) read_blocks,
				   UMBRELLA_ELF_READ_BLOCK_SIZE);
		if (retval <= 0) {
			printk(KERN_INFO "Umbrella: Unable to read signature in bloks: %d\n",retval);
			up (&umbrella_sem);
			return -EPERM;
		}
	

		lower = sh_offset;	                    /* lower bound of memset */
		upper = sh_offset + UMBRELLA_ELF_SIG_SIZE;/* upper bound */
		
		if ((lower < offset + UMBRELLA_ELF_READ_BLOCK_SIZE) &&
			(offset < upper)) {

			lower = (offset > lower) ? offset : lower;
			upper =	(offset + UMBRELLA_ELF_READ_BLOCK_SIZE <
					 upper) ? offset + UMBRELLA_ELF_READ_BLOCK_SIZE : upper;
			
			memset(read_blocks + (lower - offset), 0, upper - lower);
		}
		
		up (&umbrella_sem);

		/* continue verification loop */
		retval = umb_sign_verify_update(ctx, read_blocks, retval);
		if (retval < 0) {
			printk(KERN_INFO "Umbrella: Error updating crypto verification\n");
			return -EPERM; // <-- possible mem leak ... CHECK !!!!
		}
	}

	retval = umb_sign_verify_final(ctx, sig_result, UMBRELLA_ELF_SIG_SIZE,
								   sig_orig + UMBRELLA_BSIGN_INFOS, key_id);
	if (retval < 0) {
		return -EPERM; // <-- possible mem leak ... CHECK !!!!
	}
	
	kfree(sig_result);
 	kfree(read_blocks);
	if (ctx) {
		kfree(ctx->tvmem);
		kfree(ctx);
	}
	return retval;
	
}
/******************************************************************************
Description : 
Parameters  : 
Return value:
Comment: Replace -1 with appropiate negative return values, for more correct 
         errorhandling 
******************************************************************************/
int umb_get_pkey(unsigned char *raw_public_key_n,
				 unsigned char *raw_public_key_e, char *pkey_file) {

/*
 * Format of an MPI:
 * - 2 bytes: length of MPI in BITS
 * - MPI
 */
 	struct file *fd = NULL; 
	int i = 0;
	unsigned int len;
	unsigned char c;
	int offset = UMBRELLA_PKEY_N_OFFSET;

	fd = filp_open(pkey_file, O_RDONLY, 0400);
	if (IS_ERR(fd) || (fd == NULL)) {
		printk(KERN_INFO "Umbrella: Failed to open file\n");
		return -1;
	}

	if (!fd->f_op || !fd->f_op->mmap || !fd->f_op->read) {
		printk(KERN_INFO "Umbrella: no file operations \n");
		return -1;
	}

	kernel_read(fd, offset++, &c, 1);
	raw_public_key_n[0] = c;

	len = c << 8;

	kernel_read(fd, offset++, &c, 1);
	raw_public_key_n[1] = c;

	len |= c;
	
	len = (len + 7) / 8;

	if (len > UMBRELLA_ELF_SIG_SIZE) {
		printk(KERN_INFO "Umbrella: mpi too large \n");
		return -1;
	}

	for (i = 2; i < len+2; i++) {
		kernel_read(fd, offset++, &c, 1);
		raw_public_key_n[i] = c;
	}

	kernel_read(fd, offset++, &c, 1);
	raw_public_key_e[0] = c;
	len = c << 8;

	kernel_read(fd, offset++, &c, 1);
	raw_public_key_e[1] = c;
	len |= c;
	len = (len + 7) / 8;

	if (len > UMBRELLA_ELF_SIG_SIZE) {
		printk(KERN_INFO "Umbrella: Length of 'e' MPI is too large \n");
		return -1;
	}

	for (i = 2; i < len+2; i++) {
		kernel_read(fd, offset++, &c, 1);
		raw_public_key_e[i] = c;
	}

	filp_close(fd, 0);

	return 0;

}

/******************************************************************************
Description : 
Parameters  : 
Return value:
Comment: Replace -1 with appropiate negative return values, for more correct 
         errorhandling 
******************************************************************************/
int umb_get_pkey_from_mem(unsigned char *raw_public_key_n, 
				          unsigned char *raw_public_key_e, char *key) {

/*
 * Format of an MPI:
 * - 2 bytes: length of MPI in BITS
 * - MPI
 */
 	
	int i = 0;
	unsigned int len;
	unsigned char c;
	int offset = UMBRELLA_PKEY_N_OFFSET;
	
	c = key[offset++];
	raw_public_key_n[0] = c;
	
	len = c << 8;

	c = key[offset++];
	raw_public_key_n[1] = c;

	len |= c;	
	len = (len + 7) / 8;

	if (len > UMBRELLA_ELF_SIG_SIZE) {
		printk(KERN_INFO "Umbrella: Length of 'n' MPI is too large\n");
		return -1;
	}

	for (i = 2; i < len+2; i++) {
		c = key[offset++];
		raw_public_key_n[i] = c;
	}
	
	c = key[offset++];
	raw_public_key_e[0] = c;
	len = c << 8;

	c = key[offset++];
	raw_public_key_e[1] = c;
	
	len |= c;
	len = (len + 7) / 8;

	if (len > UMBRELLA_ELF_SIG_SIZE) {
		printk(KERN_INFO "Umbrella: Length of 'e' MPI is too large\n");
		return -1;
	}

	for (i = 2; i < len+2; i++) {
		c = key[offset++];
		raw_public_key_e[i] = c;
	}

	return 0;

}

/****************************************************************************** 
Function  : umb_parse_usig
Param:      1. char* containing the umbrella signature
Return Val: if ok -> a parse_sig struct containing the information from the 
                     signature 
            if !ok -> NULL
Comments:   The lv1 and lv2 found here can be transferred to current->security 
            using the LSM hook bprm_apply_creds. In this hook (bprm_check_security) 
            "current" is the parent, which is of no real use.
            It is the responsabillity of the caller to free the memory allocated
            for the information in the parse_sig struct and the struct itself.
******************************************************************************/

struct parsed_sig* umb_parse_usig(struct linux_binprm *bprm, char* usig  ) {

	struct parsed_sig *returnval = NULL;
	int position = 0;
	int counter,moveon,next,start,i,lv2count;
	char *tmp = NULL;

	/* variables to store the contents of the signature */
	char *signature_start = (char *)kmalloc(sizeof(char)*15,GFP_ATOMIC);
	char *id = (char *)kmalloc(sizeof(char)*UMBRELLA_ID_SIZE,GFP_ATOMIC);
	bitvector lv1 = bv_create();
	char **lv2 = (char **)kmalloc(sizeof(char*)*UMBRELLA_SIG_NO_LV2,GFP_ATOMIC);

	/* not enough memory */
	if (!signature_start && !id && !lv1 && !lv2)
		return NULL;

	if (!signature_start || !id || !lv1 || !lv2) {
		if (signature_start)
			kfree(signature_start);
		if (id)
			kfree(id);
		if (lv1)
			bv_destroy(lv1);
		if (lv2)
			kfree(lv2);
		return NULL;
	}

	
	/* parse the first part of the Umbrella signature */
	position = UMBRELLA_SIG_START+1;    
	
	for(i=0;i<15;i++)
		signature_start[i] = '\0';
	
	counter = 0;
	while( counter < 15 ) {
		if ( usig[position] != '#' ) {
			signature_start[counter] = usig[position];
		}
		counter++;
		position++;
	}
		
	if ( strcmp(signature_start,"umbrella_start") != 0 ) {
		printk(KERN_INFO "Umbrella: umb_parse_usig - not a valid umbrella header\n");
		kfree(signature_start);
		kfree(id);
		bv_destroy(lv1);
		kfree(lv2);
		return NULL; 
	}
	
	/* parse the id part of the Umbrella signature */
	position = UMBRELLA_ID_START;
	
	if ((usig[position] != 'i') || (usig[position+1] != 'd'))  {
		printk(KERN_INFO "Umbrella: 1 umb_parse_usig - no id tag found \n");
		kfree(signature_start);
		kfree(id);
		bv_destroy(lv1);
		kfree(lv2);
		return NULL;
	}
	position = position+2;
	counter = 0;
	moveon = 0;
	start = 0;
	next = 0;
	i = 0;
		
	while(!moveon && counter < UMBRELLA_ID_SIZE){
		if( usig[position] == ':' ) {
			if (!next) {
				next++;
				position++;
			} else { 
				moveon++; 
				i = position-start;
				
				/* Just to make sure nothing "extra" is in the string */
				for (counter=0;counter<i;counter++) 
					id[counter] = '\0';
				
				counter = 0;

				while ( counter<i ){ 
					if (usig[start] != ':')
						id[counter] = usig[start++];
					counter++;
				}
				id[counter] = '\0';
			} 	   					
		} else {
			if (next) 
				next--; 
			
			if (start == 0)
				start = position;
			position++;
		}
	}
	/* parse the level1 restrictions part of the Umbrella signature */ 
	position = UMBRELLA_LV1_START;
		
   	/*create a bitvector to hold the incoming lv1 restrictions */	 
	position++;
	if ((usig[position] != 'l') || (usig[position+1] != '1')) {
		printk(KERN_INFO "Umbrella: umb_parse_usig - no l1 tag found \n");
		kfree(signature_start);
		kfree(id);
		bv_destroy(lv1);
		kfree(lv2);
		return NULL;
	}
	position = position+2;
	
	i = 0;
	counter = 0;
	start = 0;
	next = 0;
	moveon = 0;
	while(!moveon){
		
		if (position >= UMBRELLA_LV2_START)
			break;
		
		if(usig[position] == ':') {
			if (next) {
				moveon++;
			} 
			
			if (start != 0) {
				counter = position-start;
				i = 0;
				tmp = (char *)kmalloc(sizeof(char* )*(counter),GFP_ATOMIC);
				while ( i<counter ){ tmp[i++] = usig[start++]; }
				tmp[i] = '\0';
				i = simple_strtol(tmp,(char **)(tmp+sizeof(tmp)),(unsigned int) 10);
				bv_bit_on(lv1,i);
				
				kfree(tmp);
				start = 0;
			}
			position++;
			next++;
			
		} else {
			if (next)  
				next--; 
			
			if (start == 0) 
				start = position;
			
			position++;
		}
	}

	/* parse the level2 part of the Umbrella signature */			
	for(i = 0;i<UMBRELLA_SIG_NO_LV2;i++)
		lv2[i] = NULL;
	
	position = UMBRELLA_LV2_START;
	
	position++;
	if ( ( usig[position] != 'l' ) && ( usig[position+1] != '2' ) )  {
		printk(KERN_INFO "Umbrella: umb_parse_usig - no l2 tag found \n");
		kfree(signature_start);
		kfree(id);
		bv_destroy(lv1);
		kfree(lv2);
		return NULL;
	}
	position = position+2;
	
	i = 0;
	counter = 0;
	start = 0;
	next = 0;
	moveon = 0;
	lv2count = 0;
	while(!moveon){
		if(usig[position] == ':') {
			if (next) {
				moveon++;
			} 
			
			if (start != 0) {
				i = 0;
				counter = position-start;
				tmp = (char *)kmalloc(sizeof(char)*(counter),GFP_ATOMIC);
				while ( i<counter ){ tmp[i++] = usig[start++]; }
				tmp[i] = '\0';
				lv2[lv2count] = tmp;
				start = 0;
				lv2count++;
			}
			position++;
			next++;
			
		} else {
			if (next)  
				next--; 
				
			if (start == 0) 
				start = position;
			
			position++;
		}
	}
	
	/* Finally we allocate space and assigns values to the return type */
	returnval = kmalloc(sizeof(struct parse_sig*),GFP_ATOMIC);
	if (!returnval) {
		printk(KERN_INFO "Umbrella: Failed allocating memory for "
                         " structure in umb_parse_usig!\n");
		return NULL;
	}
	
	returnval->sig_start = signature_start;
	returnval->id = id;
    returnval->lv1 = lv1;
	returnval->lv2 = lv2;
	return returnval;
}

/******************************************************************************
Function:   umb_to_task_struct
Parameters: 

Comments:    
******************************************************************************/
void umb_to_task_struct(struct security_struct *security, struct parsed_sig *p ){

	char **level2;
	int i;

	if (p->lv1) {
		bv_or((bitvector)security->bin_level1, 
			  (bitvector)security->bin_level1,
			  (bitvector)p->lv1); 
	} 

	/* Repeat for binary level 2 restrictions */

	if (p->lv2) {
		level2 = p->lv2;
		if(level2 && *level2 != NULL) {
			i = 0;
			while ((level2[i] != NULL) && (i < UMBRELLA_SIG_NO_LV2)) {
				security->bin_level2[i] = (char *)kmalloc(sizeof(char) * 
														  (strlen(level2[i]) + 1),
														  GFP_ATOMIC);
				strcpy(security->bin_level2[i], level2[i]);
				i++;
			}
		}
	} 

}

/******************************************************************************
Function:   umb_to_task_struct
Parameters: 

Comments:    
******************************************************************************/
int umb_cache_signature(struct file *f, struct parsed_sig *p) {
	
	struct inode *inode = f->f_dentry->d_inode;

	if( inode != NULL ) {
		if( inode->i_security == NULL ) {
			inode->i_security = p;
			return 0;
		} else { 
			return -1; 
		}
	} else { 
		return -1; 
	}
}


/******************************************************************************
Function:   umb_handle_signature
Parameters: 1. A linux_binprm struct that holds information about the binary. 
               Look in include/linux/binfmts.h for more details
Return Val: if ok -> 0
            if !ok -> -EPERM
            if error -> -1 

Comments:   This is the entry point for the signature part of Umbrella and is 
            called from the bprm_check_security hook. All other functions in 
            this file, will only be envoked as a result of actions coming from 
            here. 
******************************************************************************/

int umb_handle_signature(struct linux_binprm * bprm) {

	struct security_struct *security = current->security;
	struct parsed_sig *parsed_usig = NULL;
	struct elfhdr *elf_hdr;
	Elf32_Shdr *elf_shdata;
	int size = 0;
	int signature_offset = 0;
	struct dsig *file_sig;
	struct file *file = bprm->file;
	int retval = 0;
	
	 /* This should NOT happen according to the LSM documentation, but anyway lets 
		check just to be sure! */
	if ( file == NULL )
		return -EPERM; /* is this the correct action here I wonder ?? */
	
	elf_hdr = umb_read_elf_header(file);
	if (elf_hdr != NULL) {
		size = elf_hdr->e_shnum * sizeof(Elf32_Shdr);
		elf_shdata = umb_read_section_header(file, size, elf_hdr->e_shoff);
		if (elf_shdata == NULL) {
			printk(KERN_INFO "Umbrella: Unable to find elf_shdata\n ");
			kfree(elf_hdr);
			return -EPERM;
		}

		/* Let's find the signature section */
		file_sig = umb_find_signature(elf_hdr, elf_shdata, file, &signature_offset);
		if (!file_sig) {
			kfree(elf_hdr);
			kfree(elf_shdata);

/* If Umbrella is compiled in DEBUG mode, files without signatures
   will be permitted to run, because 0 is returned. If in normal
   mode return -EPERM to prevent execution */
#if defined(CONFIG_SECURITY_UMBRELLA_DEBUG) 
			return 0;       
#else
			return -EPERM; 
#endif
		}

		/* now lets parse the usig to extract the lv1 and lv2 restrictions */
		/* and the id of the key it was signed with */

		parsed_usig = umb_parse_usig(bprm,file_sig->usig);
 
		if (parsed_usig == NULL) {
			printk(KERN_INFO "Umbrella: Error parsing the Umbrella Signature\n ");
			return -EPERM;
		}
		
		/* good .. now lets verify the signature */
		retval = umb_verify_signature(elf_shdata, file_sig->bsig, file,
									  signature_offset, parsed_usig->id);

		
		/* the signature was good, let's check for a publickey to insert, */
		/* and apply execute restrictions to the process */
 		if ( retval == 0 ) {
			
			/* If the file executed is a "insert_VENDORTAG_key" file */
			/* then we insert the key into the key ring :-)          */
			if(file_sig->key) 
				umb_key_insert(file_sig->key, file_sig->keyid);

			if (security != NULL) {

 				/* Put the binary restrictions in the task
				   security field for use in task_alloc security. 
				   Remember to free the allocated memory in task_alloc
				*/

				umb_to_task_struct(security,parsed_usig);
				
				/* Cache the restrictions in the
				   file->f_dentry->d_inode->i_security field to indicate that
				   the signature was successfully authenticated
				*/

				printk("MTDEBUG: before signature cache\n");
				if ( umb_cache_signature(file,parsed_usig) ) {
					printk("Umbrella: Signature caching failed for  %s\n",
						   file->f_dentry->d_name.name);
					retval = -EPERM;
				}

			}
		} else if (retval > 0) {
			printk(KERN_INFO "Umbrella: Signature do not match for %s\n",
				   file->f_dentry->d_name.name);
			retval = -EPERM;
		} else {
			printk(KERN_INFO "Umbrella: Signature verification failed because of errors: %d for %s \n",
				   retval, file->f_dentry->d_name.name);
			retval = -EPERM;
		}

		/* lets free the stuff we no longer need */
		kfree(elf_hdr);
		kfree(elf_shdata);
		kfree(file_sig->bsig);
		kfree(file_sig->usig);
		kfree(file_sig);
/* 		if (parsed_usig != NULL) */
/*  			umb_free_sig(parsed_usig); */
	}  
	
	return retval; 
}

/******************************************************************************
Function:   umb_verify_elf
Parameters: 1. elf_hdr is an elf header
Return Val: if ok -> 0
            if !ok -> -1  
Comments:   Just a basic check of the elf header to verify the correctness.
******************************************************************************/
int umb_verify_elf(struct elfhdr *elf_hdr) {
	if (memcmp(elf_hdr->e_ident, ELFMAG, SELFMAG) != 0) {
		return -1;
	}
	
	if (!elf_hdr->e_shoff) {
		printk(KERN_INFO "Umbrella: No section header!\n");
		return -1;
	}
	
	if (elf_hdr->e_shentsize != sizeof(Elf32_Shdr)) {
		printk(KERN_INFO "Umbrella: Section header is wrong size!\n");
		return -1;
	}
	
	if (elf_hdr->e_shnum > 65536U / sizeof(Elf32_Shdr)) {
		printk(KERN_INFO "Umbrella: Too many entries in Section Header!\n");
		return -1;
	}

	return 0;
}

/******************************************************************************
Function:   umb_read_elf_header
Parameters: 1. The file struct from where the elf_header should be read
Return Val: if ok -> a pointer to the elfheader 
            if !ok -> NULL
Comments:   
******************************************************************************/

struct elfhdr *umb_read_elf_header(struct file *file) {
	struct elfhdr *elf_hdr;

	elf_hdr = (struct elfhdr *)kmalloc(sizeof(struct elfhdr), GFP_ATOMIC);
	if (!elf_hdr) {
		printk(KERN_INFO "Umbrella: kmalloc failed for elf_ex\n");
		return NULL;
	}
	
	if (file != NULL)
		kernel_read(file, 0, (char *)elf_hdr, sizeof(struct elfhdr));


	if (umb_verify_elf(elf_hdr)) {
		kfree(elf_hdr);
		return NULL;
	}

	return elf_hdr;
}

Elf32_Shdr *umb_read_section_header(struct file *file, int sh_size, int sh_off) {
	
	Elf32_Shdr *elf_shdata;
	int retval;
	
	elf_shdata = (Elf32_Shdr *) kmalloc(sh_size, GFP_ATOMIC);
	if (!elf_shdata) {
		printk(KERN_INFO "Umbrella: read_section_header - cannot allocate memory to read Section Header\n");
		return NULL;
	}
	
	retval = kernel_read(file, sh_off, (char *)elf_shdata, sh_size);

	if (retval < 0 || retval != sh_size) {
		printk(KERN_INFO "Umbrella: Unable to read binary %s: %d\n", file->f_dentry->d_name.name, retval);
		kfree(elf_shdata);
		return NULL;
	}
	return elf_shdata;
}

/******************************************************************************
Function:   umb_find_signature
Parameters: 1. elf_hdr is an elf header
            2. elf_shdata holds all entries in the section header table of the elf.
            3. file is the file handler of the binary.
            4. sh_offset is the offset of  the signature section in the elf and is 
               set to offset the siganture section in the elf file. 

Return Val: if ok -> a buffer containing the signature in the elf header
            if !ok -> NULL
Comments:   caller is responsible for deallocating the buffer returned.
******************************************************************************/

struct dsig* umb_find_signature(struct elfhdr *elf_hdr, Elf32_Shdr * elf_shdata, 
								struct file *file,int *sh_offset) {
	int bsign_entry, i;
	int retval = 0;
	int retval2 = 0; /* Hack alert */
	struct dsig *dsig = NULL;
	int keylength = 0;
	char tmp[UMB_KEY_ENDTAG_SIZE];

	bsign_entry = 0;
	while (bsign_entry < elf_hdr->e_shnum &&
		   elf_shdata[bsign_entry].sh_type != UMBRELLA_ELF_SIG_SECTION) {
		bsign_entry++;
	}

	if (bsign_entry == elf_hdr->e_shnum)
		return NULL;
	
	if (elf_shdata[bsign_entry].sh_type == UMBRELLA_ELF_SIG_SECTION) {
	
		/* Now get BSIGN_ELF_SIG_SIZE bytes of signature section */

		dsig = (struct dsig*) kmalloc(sizeof(struct dsig), GFP_ATOMIC);
		if (!dsig) {
			printk(KERN_INFO "Umbrella: Unable to alloc return signature struct\n");
			return NULL;
		}

		if (elf_shdata[bsign_entry].sh_size != UMBRELLA_ELF_SIG_SIZE) {
			printk(KERN_INFO "Umbrella: Signature section is not %u bytes\n",
				   UMBRELLA_ELF_SIG_SECTION);
			return NULL;
		}
				
		dsig->bsig = (char *) kmalloc(UMBRELLA_ELF_SIG_SIZE, GFP_ATOMIC);
		if (!dsig->bsig) {
			printk(KERN_INFO "Umbrella: umb_signature.c: kmalloc failed for bsig buffer.\n");
			kfree(dsig);
			return NULL;
		}

		retval = kernel_read(file, elf_shdata[bsign_entry].sh_offset,(char *) 
							 dsig->bsig, UMBRELLA_ELF_SIG_SIZE);

		dsig->usig = (char*) kmalloc(UMBRELLA_SIG_SIZE, GFP_ATOMIC);	
		if (!dsig->usig) {
			printk(KERN_INFO "Umbrella: umb_signature.c: kmalloc failed for usig buffer.\n");
			kfree(dsig->bsig);
			kfree(dsig);
			return NULL;
		}

		/* ok, now lets get the Umbrella signature by calculating the offset by
 		   subtracting the umbrella signature size from the bsign offset */

		retval2 = kernel_read(file, elf_shdata[bsign_entry].sh_offset -
							  UMBRELLA_SIG_SIZE, (char *) dsig->usig,
							  UMBRELLA_SIG_SIZE);

		for (i=0;i<UMB_KEY_ENDTAG_SIZE;i++)
			tmp[i] = '\0';

	    /* READ END KEY TAG */
		kernel_read(file, elf_shdata[bsign_entry].sh_offset-
					UMBRELLA_SIG_SIZE   - UMB_KEY_SIZE_SIZE-
					UMB_KEY_ID_SIZE - UMB_KEY_ENDTAG_SIZE,
					tmp, UMB_KEY_ENDTAG_SIZE);

		if (!strcmp(tmp, "#umbkeyend#")) {

			for (i=0;i<UMB_KEY_ENDTAG_SIZE;i++)
				tmp[i] = '\0';

			/* READ KEY SIZE */
			kernel_read(file, elf_shdata[bsign_entry].sh_offset-UMBRELLA_SIG_SIZE-
						UMB_KEY_SIZE_SIZE, tmp, UMB_KEY_SIZE_SIZE);

			keylength = simple_strtol(tmp, (char **)NULL, 10);

			dsig->key = (char *) kmalloc(UMB_KEY_STARTTAG_SIZE + keylength +
										 UMB_KEY_ENDTAG_SIZE   + UMB_KEY_SIZE_SIZE,
										 GFP_ATOMIC);

			dsig->keyid = (char *) kmalloc(UMB_KEY_ID_SIZE, GFP_ATOMIC);

			if (!dsig->key || !dsig->keyid) {
				printk(KERN_INFO "Umbrella: umb_signature.c: kmalloc failed for usig buffer.\n");
				kfree(dsig->bsig);
				kfree(dsig->usig);
				kfree(dsig);
				return NULL;
			}

			/* READ KEYID */
			kernel_read(file, elf_shdata[bsign_entry].sh_offset     -
						UMBRELLA_SIG_SIZE   - UMB_KEY_SIZE_SIZE     -
						UMB_KEY_ID_SIZE, dsig->keyid, UMB_KEY_ID_SIZE);

			for(i=0;i<UMB_KEY_ID_SIZE;i++) {
				if(dsig->keyid[i] == ':')
					dsig->keyid[i] = '\0';
			}
			
			/* READ COMPLETE KEY SECTION */
			kernel_read(file, elf_shdata[bsign_entry].sh_offset     -
						UMBRELLA_SIG_SIZE   - UMB_KEY_SIZE_SIZE     -
						UMB_KEY_ID_SIZE     - UMB_KEY_ENDTAG_SIZE   -
						keylength, dsig->key, keylength);
		}
		else
			dsig->key = NULL;


		if ((retval < 0) || (retval != UMBRELLA_ELF_SIG_SIZE) || ( retval2 < 0 )) {
			printk(KERN_INFO "Umbrella: Unable to read signature: %d\n", retval);
			kfree(dsig->bsig);
			kfree(dsig->usig);
			kfree(dsig->key);
			kfree(dsig);
			return NULL;
		}

		*sh_offset = elf_shdata[bsign_entry].sh_offset;
	}
	
	return dsig;
}

/******************************************************************************
Function:   umb_free_sig
Parameters: 1. the parsed usig to be freed
Return Val: 
Comments:   
******************************************************************************/
void umb_free_sig(struct parsed_sig* usig) {
	int i=0;

	kfree(usig->sig_start);
	kfree(usig->id);
	bv_destroy(usig->lv1);
	while( usig->lv2[i] != NULL ) {
		kfree(usig->lv2[i]);
		i++;
	}
	kfree(usig);
}

/****************************************************************************** 
 Function :  umb_test_sig
   Param:      1. a char* containing the umbrella signature
   Return Val: void 
   Commets: Prints the begin position of the umbrella specific parts of the 
            signature
******************************************************************************/

void umb_test_sig(char* usig){
	int i = 0;
	
	while(i<UMBRELLA_SIG_SIZE){
		if (usig[i] == '#')
			printk(KERN_INFO "Umbrella: position was %i and a # was found \n",i);         
		i++;
	}
}

/****************************************************************************** 
 Function :  umb_cached_signature
   Param:      1. bprm struct containing the file with the cached decision
   Return Val: 0 if good -1 if problem
   Commets: 
******************************************************************************/

int umb_cached_signature(struct linux_binprm * bprm) {

	struct file *f = bprm->file;
	struct parsed_sig *p = NULL;
	struct security_struct *security = current->security;
	struct inode *inode = f->f_dentry->d_inode;

	if ( f->f_dentry->d_inode->i_security != NULL ) {
 		p = (struct parsed_sig *) inode->i_security;
		umb_to_task_struct(security,p);
		return 0;
	} else {
		return -1;	
	}
}
