#!/usr/bin/env ruby

# $Header: /cvsroot/umbrella/umbrella-devel/libumbrella/scripts/sign_file.rb,v 1.4 2005/03/02 14:06:57 ipqw Exp $


if not VERSION =~ /1\.8/
  puts "This script needs Ruby >=1.8"
  exit
end

trap(2) { puts "\nexiting ..."; exit 1 }

# definitions of global constants
BEGIN_TAG = "#umbrella_start#"
ID_TAG = "#id:"
L1_TAG = "#l1:"
L2_TAG = "#l2:"
END_TAG = "#umbrella_end#"

UMBRELLA_SECTION_SIZE = 2346
VENDOR_ID_SIZE = 64
L1_SIZE = 192
L2_SIZE = 2048


def display_help
  puts
  puts "Usage for signing file:    #{$0} <--id> <--l1> <--l2> --file=file"
  puts "Display this help message: #{$0} <--help|-h>"
  puts
  puts "Example: #{$0} --id=UmbrellaInc --l1=11:25 --l2=/secret:/verysecret --file=ProgramToBeSigned"
  puts
  puts "Note: * The --id is an optimization for future versions. At the moment, the option can be omitted."
  puts "      * Speficy only ONE file (do not use wildcards)"
  puts
  puts "Copyright (c) 2004, 2005 The Umbrella Team - http://umbrella.sourceforge.net"
  puts
end


def pad_l1 l1_restrictions
  padding_size = L1_SIZE - l1_restrictions.length

  padding = ""
  padding_size.times {
    padding = padding + ":"
  }
  
  final_l1 = l1_restrictions + padding

  return final_l1
end


def pad_l2 l2_restrictions
  padding_size = L2_SIZE - l2_restrictions.length  

  padding = ""
  padding_size.times {
    padding = padding + ":"
  }
  
  final_l2 = l2_restrictions + padding

  return final_l2
end


def pad_vendor_id vendor_id
  padding_size = VENDOR_ID_SIZE - vendor_id.length  

  padding = ""
  padding_size.times {
    padding = padding + ":"
  }
  
  final_id = vendor_id + padding

  return final_id
end

class Arguments
  # set methods
  def set_vendor_id(id)
    @id = id
  end

  def set_level1(l1)
    @l1 = l1
  end

  def set_level2(l2)
    @l2 = l2
  end

  def set_file(file)
    @file = file
  end

  # get methods
  def get_vendor_id
    @id
  end

  def get_level1
    @l1
  end

  def get_level2
    @l2
  end

  def get_file
    @file
  end
end


def get_input argv
  arguments = Arguments.new
  argv.each { |input|
    if input =~ /^--id=/
      id = $'
      # match only word characters: A-Za-z0-9_
      if not id =~ /^\w+$/
        puts "*** ERROR: Invalid ID. Only the characters A-Z, a-z, 0-9 and _ are allowed"
        puts "aborting ..."
        exit 1
      end
###      puts "Setting vendor ID to " + id
      arguments.set_vendor_id(id)
    elsif input =~ /^--l1=/
      l1 = $'
      # match (1 digit 1 optional digit, optional colon)-minimum one time
      if not l1 =~ /^(\d\d?:?)+$/
        puts "*** ERROR: Invalid level 1 restrictions. Only the characters 0-9 and : are allowed"
        puts "aborting ..."
        exit 1
      end
      # remove tailing : if present
      l1 = l1.chop if l1 =~ /:$/
###      puts "Setting Level 1 to " + l1
      arguments.set_level1(l1)
    elsif input =~ /^--l2=/
      l2 = $'
      # remove tailing : if present
      l2 = l2.chop if l2 =~ /:$/
###      puts "Setting Level 2 to " + l2
      arguments.set_level2(l2)
    elsif input =~ /^--file=/
###      puts "Setting file to " + $'
      arguments.set_file($')
    else
      puts "Could not parse invalid argument: #{input}"
    end
  }

  if arguments.get_file == nil
    puts "*** ERROR: No file specified... exiting"
    puts "aborting ..."
    exit 1
  end

  if arguments.get_level1 == nil and arguments.get_level2 == nil
    puts "*** WARNING: Signing file without any restrictions (ok)"
  end

  # fix up non-existing strings :)
  arguments.set_vendor_id("") if arguments.get_vendor_id == nil
  arguments.set_level1("") if arguments.get_level1 == nil
  arguments.set_level2("") if arguments.get_level2 == nil

  return arguments
end




########### MAIN ############

# check correct number of command line arguments or help
if ARGV.length == 0 or ARGV.at(0) =~ /(^-h|^--help)/ or ARGV.length > 4
  display_help
  exit 1
end

arguments = get_input(ARGV)

# check if file exists
if not test(?f, arguments.get_file)
  puts "*** ERROR: File does not exist"
  puts "aborting ..."
  exit 1
end

# check if the file is an ELF binary
`readelf -S #{arguments.get_file} >/dev/null 2>&1`
if $? != 0
  puts "*** ERROR: The file supplied is NOT an ELF binary!"
  puts "aborting ..."
  exit 1
end

# check if bsign is present
`bsign >/dev/null 2>&1`
if $? == 127
  puts "*** ERROR: Bsign is not available through your PATH environment variable!"
  puts "Bsign can be downloaded from: http://packages.debian.org/unstable/admin/bsign.html"
  puts "aborting ..."
  exit 1
end

# create Umbrella section
umbrella_section = BEGIN_TAG + ID_TAG + pad_vendor_id(arguments.get_vendor_id) +
                               L1_TAG + pad_l1(arguments.get_level1)  +
                               L2_TAG + pad_l2(arguments.get_level2)  +  END_TAG

# be SURE that the size of the Umbrella_section is OK
if umbrella_section.length != UMBRELLA_SECTION_SIZE
  puts "*** ERROR: Umbrella section has invalid size!!"
  puts "Section length was #{umbrella_section.length} but should be #{UMBRELLA_SECTION_SIZE}"
  puts "aborting ..."
  exit 1
end

# append the Umbrella section to file
elf_file = File.open(arguments.get_file, "a") # a=append (write only, seek end of file)
elf_file << umbrella_section
elf_file.close

puts `bsign -s #{arguments.get_file}`
if $? != 0
  puts "*** ERROR: Error signing file!!"
  puts "Incorrect passphrase provided or GPG not available. Be aware that the \"Umbrella-part\" IS allready appended to the file. To sign the file manually, type: bsign -s #{arguments.get_file}"
  puts "aborting ..."
  exit 1
end

puts "File #{arguments.get_file} signed successfully."

exit 0
